import requests
from amdirt.core import logger
import os
from typing import List, Dict


class ENA:
    """Class for querying the ENA API"""

    def __init__(self) -> None:
        self.base_url = "https://www.ebi.ac.uk/ena"

    def __repr__(self) -> str:
        """Display URL of API documentation"""
        return f"The documentation for the {self.__class__.__name__} API can be found at {self.base_url}"

    def status(self) -> bool:
        """Check if API is up

        Returns:
            bool: True if API is up, False otherwise
        """
        try:
            resp = requests.get(self.base_url)
            if resp.status_code == 200:
                return True
            else:
                return False
        except requests.exceptions.ConnectionError:
            return False

    def doc(self, dir: str = ".") -> None:
        """Get PDF documentation for API

        Args:
            dir(str): path to output PDF directory
        """
        r = requests.get(self.base_url + "doc")
        pdf = os.path.join(
            os.path.expanduser(dir), f"{self.__class__.__name__}_APIDocumentation.pdf"
        )
        with open(pdf, "wb") as fw:
            fw.write(r.content)
        logger.info(
            f"{self.__class__.__name__} documentation has been written to {pdf}"
        )

    def __get_json__(self, url: str) -> List[Dict]:
        """Get json content from URL

        Args:
            url(str): URL to get json content from
        Returns:
            List[Dict]: json content
        """
        resp = requests.get(url)
        if resp.status_code == 200:
            if len(resp.json()) > 0:
                return resp.json()
            else:
                logger.warning("No results found")
                return []


class ENABrowserAPI(ENA):
    """Class to interact with the ENA Browser API"""

    def __init__(self) -> None:
        super().__init__()
        self.base_url = "https://www.ebi.ac.uk/ena/browser/api/"


class ENAPortalAPI(ENA):
    """Class to interact with the ENA Portal API"""

    def __init__(self) -> None:
        """ENA Portal API class

        Attributes:
            base_url(str): base URL for ENA Portal API
        """
        super().__init__()
        self.base_url = "https://www.ebi.ac.uk/ena/portal/api/"

    def _get_results(self) -> List[Dict]:
        """Get list of available results

        Returns:
            List[Dict]: list of available results
        """
        url = os.path.join(self.base_url, "results?dataPortal=ena&format=json")
        json_resp = self.__get_json__(url)
        return json_resp

    def list_results(self) -> None:
        """Display list of available results

        Returns:
            List[Dict]: list of available results
        """
        json_resp = self._get_results()
        for result in json_resp:
            logger.info(f"{result['resultId']} - {result['description']}")
        return json_resp

    def _get_fields(self, result_type: str) -> List:
        """Get list of available fields

        Args:
            result_type(str): A result is a set of data that can
            be searched against and returned
        Returns:
            List: list of available fields
        """
        url = os.path.join(
            self.base_url,
            f"returnFields?dataPortal=ena&format=json&result={result_type}",
        )
        json_resp = self.__get_json__(url)
        return json_resp

    def list_fields(self, result_type: str) -> None:
        """Display list of available fields

        Args:
            result_type(str): A result is a set of data that can
            be searched against and returned
        Returns:
            List: list of available fields
        """
        json_resp = self._get_fields(result_type=result_type)
        for field in json_resp:
            logger.info(f"{field['columnId']} - {field['description']}")
        logger.info(f"Available fields for {result_type} are: {json_resp}")

    def _check_result_type(self, result_type: str) -> bool:
        """Check if result type is allowed

        Args:
            result_type(str): A result is a set of data that can
            be searched against and returned
        Returns:
            bool: True if result type is valid, False otherwise
        """
        all_results = self._get_results()
        results = [result["resultId"] for result in all_results]
        if result_type not in results:
            logger.warning(f"{result_type} is not a valid result type")
            return False
        return True

    def _check_fields(self, result_type: str, fields: List[str]) -> bool:
        """Check if fields are allowed

        Args:
            result_type(str): A result is a set of data that can
            be searched against and returned
            fields(List): list of fields to check
        Returns:
            bool: True if fields are valid, False otherwise
        """
        all_fields = self._get_fields(result_type)
        fields = [field["columnId"] for field in all_fields]
        for field in fields:
            if field not in fields:
                logger.warning(f"{field} is not a valid field")
                return False
        return True

    def query(
        self,
        accession: str,
        result_type: str = "read_run",
        fields: List = [
            "run_accession",
            "sample_accession",
            "fastq_ftp",
            "fastq_md5",
            "fastq_bytes",
        ],
    ) -> dict:
        """Generate list of runs metadata for a study accession

        Args:
            accession (str): ENA accession
            result_type(str): A result is a set of data that can
            be searched against and returned
            fields(List): list of fields to return

        Returns:
            dict: run_accession as keys, and metadata as values
        """

        self._check_result_type(result_type)
        self._check_fields(result_type, fields)
        url = os.path.join(
            self.base_url,
            f"filereport?accession={accession}&download=false&format=json&result={result_type}&fields={','.join(fields)}",
        )
        json_resp = self.__get_json__(url)
        return json_resp


if __name__ == "__main__":
    e = ENAPortalAPI()
    print(e.query("PRJEB30331"))



from typing import Iterable, AnyStr, Union
from amdirt.validate.domain import DatasetValidator, DFError
from amdirt.core import get_json_path
from amdirt.core.diff import get_sample_diff
from amdirt.core.ena import ENAPortalAPI
from rich.progress import track
from pathlib import Path
import pandas as pd
import json


class AMDirValidator(DatasetValidator):
    """Validator Class for AncientMetagenomeDir datasets"""

    def check_duplicate_dois(self) -> bool:
        project_dois = self.dataset.groupby("project_name")[
            "publication_doi"
        ].unique()
        doi_unique = self.dataset.groupby("project_name")[
            "publication_doi"
        ].nunique()
        err_cnt = 0
        for project in doi_unique.index:
            if doi_unique[project] > 1:
                err_cnt += 1
                self.add_error(
                    DFError(
                        error="Duplicated DOI Error",
                        source=project_dois[project],
                        column="publication_doi",
                        row="",
                        message=f"Duplicate DOI for {project} project. Make sure each project has a single DOI",
                    )
                )
        if err_cnt > 0:
            return False
        return True

    def check_multi_values(
        self, column_names: Iterable[str] = ["archive_accession"]
    ) -> bool:
        """Check for duplicates entries in multi values column
        Args:
            column_names (Iterable[str], optional): List of multi values columns to check for duplications. Defaults to ["archive_accession"].
        """
        err_cnt = 0
        for column in column_names:
            row = 0
            for archives in self.dataset[column]:
                archives = archives.split(",")
                if len(set(archives)) != len(archives):
                    self.add_error(
                        DFError(
                            error="Duplicates in multi values column",
                            source=archives,
                            column=column,
                            row=row,
                            message=f"Duplicates in multi values column {column}. Make sure each value in combination is unique",
                        )
                    )
                    err_cnt += 1
                row += 1
        if err_cnt > 0:
            return False
        else:
            return True

    def check_sample_accession(self, remote: Union[AnyStr, None] = None) -> bool:
        """Check that sample accession are valid

        Args:
            remote (AnyStr | None, optional): Remote to check against. Defaults to None.
        """
        if not remote:
            with open(get_json_path()) as f:
                tables = json.load(f)
            samples = tables["samples"]
            for table in samples:
                if self.dataset_name == Path(samples[table]).name:
                    remote = samples[table]
            if remote is None:
                raise SystemExit(
                    f"No remote found for {self.dataset} dataset, please provide one"
                )
        remote_samples = DatasetValidator(schema=self.schema_path, dataset=remote)
        df_change = pd.concat([remote_samples.dataset, self.dataset]).drop_duplicates(
            keep=False
        )
        df_change.drop_duplicates(
            inplace=True, keep="last", subset=list(df_change.columns)[:-1]
        )
        is_ok = True
        
        print(df_change)

        if df_change.shape[0] > 0:
            e = ENAPortalAPI()
            change_dict = {}

            for i in df_change.index:
                try:
                    supported_archive = df_change.loc[i, "archive"] in ["SRA", "ENA"]
                except ValueError as e:
                    print(e)
                    print(df_change.loc[i, :])
                    supported_archive = False
                    continue
                if supported_archive:
                    samples = df_change.loc[i, "archive_accession"].split(",")
                    project = df_change.loc[i, "archive_project"]
                    if project not in change_dict:
                        change_dict[project] = {"index": i, "sample": samples}
                    else:
                        change_dict[project]["sample"].extend(samples)
                else:
                    continue

            for project in track(change_dict, description="Checking ENA/SRA accessions..."):
                json_result = e.query(
                    accession=project,
                    result_type="read_experiment",
                    fields=["secondary_sample_accession"],
                )
                ena_samples = []
                for i in json_result:
                    ena_samples.append(i["secondary_sample_accession"])
                for sample in change_dict[project]["sample"]:
                    if sample not in ena_samples:
                        row = df_change.query(
                            f"archive_accession.str.contains('{sample}') and archive_project.str.contains('{project}')"
                        ).index[0]
                        self.add_error(
                            DFError(
                                error="Invalid sample accession",
                                source=sample,
                                column="archive_accession",
                                row=row,
                                message=f"Sample accession {sample} is not a valid ENA/SRA sample accession for the project {project}",
                            )
                        )
                        is_ok = False
        return is_ok



import streamlit as st
import pandas as pd
import os
from st_aggrid import GridOptionsBuilder, AgGrid, GridUpdateMode, DataReturnMode, JsCode
import argparse
import zipfile
import json
import os
from amdirt import __version__
from amdirt.core import (
    prepare_bibtex_file,
    prepare_eager_table,
    prepare_mag_table,
    prepare_accession_table,
    prepare_aMeta_table,
    prepare_taxprofiler_table,
    is_merge_size_zero,
    get_amdir_tags,
    get_libraries,
    get_json_path,
)


st.set_page_config(
    page_title="amdirt viewer",
    page_icon="https://raw.githubusercontent.com/SPAAM-community/AncientMetagenomeDir/master/assets/images/logos/spaam-AncientMetagenomeDir_logo_mini.png",
    layout="wide",
)

supported_archives = ["ENA", "SRA"]

if "compute" not in st.session_state:
    st.session_state.compute = False
if "force_samp_validation" not in st.session_state:
    st.session_state.force_samp_validation = False
if "force_lib_validation" not in st.session_state:
    st.session_state.force_lib_validation = False
if "table_name" not in st.session_state:
    st.session_state.table_name = None


def parse_args():
    parser = argparse.ArgumentParser("Run Streamlit app")
    parser.add_argument(
        "-c", "--config", help="json config file", default=get_json_path()
    )
    try:
        args = parser.parse_args()
    except SystemExit as e:
        os._exit(e.code)
    return args


args = parse_args()

tags = get_amdir_tags() + ["master"]

with open(args.config) as c:
    tables = json.load(c)
    samples = tables["samples"]
    libraries = tables["libraries"]

# Sidebar
with st.sidebar:
    st.markdown(
        """
<p style="text-align:center;"><img src="https://raw.githubusercontent.com/SPAAM-community/AncientMetagenomeDir/master/assets/images/logos/spaam-AncientMetagenomeDir_logo_colourmode.svg" alt="logo" width="50%"></p>
""",
        unsafe_allow_html=True,
    )
    st.write(f"# [amdirt](https://github.com/SPAAM-community/amdirt) viewer tool")
    st.write(f"\n Version: {__version__}")
    st.session_state.tag_name = st.selectbox(
        label="Select an AncientMetagenomeDir release", options=tags
    )
    options = ["No table selected"] + list(samples.keys())
    st.session_state.table_name = st.selectbox(label="Select a table", options=options)
    st.session_state.height = st.selectbox(
        "Number of rows to display", (10, 20, 50, 100, 200), index=1
    )
    st.session_state.dl_method = st.selectbox(
        label="Data download method",
        options=["curl", "nf-core/fetchngs", "aspera", "sratookit"],
    )
    if st.session_state.dl_method == "aspera":
        st.warning(
            "You will need to set the `${ASPERA_PATH}` environment variable. See [documentation](https://amdirt.readthedocs.io) for more information."
        )
    st.warning(
        f"Only {' and '.join(supported_archives)} archives are supported for now"
    )

if st.session_state.table_name != "No table selected":
    # Main content
    st.markdown(f"AncientMetagenomeDir release: `{st.session_state.tag_name}`")
    st.markdown(f"Displayed table: `{st.session_state.table_name}`")
    samp_url = samples[st.session_state.table_name].replace(
        "master", st.session_state.tag_name
    )
    lib_url = libraries[st.session_state.table_name].replace(
        "master", st.session_state.tag_name
    )
    df = pd.read_csv(
        samp_url,
        sep="\t",
    )
    library = pd.read_csv(
        lib_url,
        sep="\t",
    )
    gbs = GridOptionsBuilder.from_dataframe(df)
    gbs.configure_default_column(
        groupable=True,
        value=True,
        enableRowGroup=True,
        aggFunc="sum",
        editable=False,
        filterParams={"inRangeInclusive": "true"},
    )
    gbs.configure_selection(selection_mode="multiple", use_checkbox=True)
    gbs.configure_grid_options(checkboxSelection=True)

    gbs.configure_pagination(
        enabled=True,
        paginationAutoPageSize=False,
        paginationPageSize=st.session_state.height,
    )
    gbs.configure_column(
        "project_name",
        headerCheckboxSelection=True,
        headerCheckboxSelectionFilteredOnly=True,
    )
    gridOptions_sample = gbs.build()

    gbl = GridOptionsBuilder.from_dataframe(library)
    gbl.configure_default_column(
        groupable=True,
        value=True,
        enableRowGroup=True,
        aggFunc="sum",
        editable=False,
        filterParams={"inRangeInclusive": "true"},
    )
    gbl.configure_selection(selection_mode="multiple", use_checkbox=True)
    gbl.configure_grid_options(checkboxSelection=True)

    gbl.configure_pagination(
        enabled=True,
        paginationAutoPageSize=False,
        paginationPageSize=st.session_state.height,
    )
    gbl.configure_column(
        "project_name",
        headerCheckboxSelection=True,
        headerCheckboxSelectionFilteredOnly=True,
    )
    gridOptions_library = gbl.build()

    with st.form("Samples table") as f:
        st.markdown("Select samples to filter")
        df_mod = AgGrid(
            df,
            gridOptions=gridOptions_sample,
            data_return_mode="filtered",
            update_mode="selection_changed",
        )
        if st.form_submit_button("Validate sample selection", type="primary"):
            if len(df_mod["selected_rows"]) == 0:
                st.error(
                    "You didn't select any sample! Please select at least one sample."
                )
            else:
                st.session_state.compute = True

    merge_is_zero = is_merge_size_zero(
        pd.DataFrame(df_mod["selected_rows"]), library, st.session_state.table_name
    )

    if (
        st.session_state.compute
        and not merge_is_zero
        and pd.DataFrame(df_mod["selected_rows"]).shape[0] != 0
    ):
        nb_sel_samples = pd.DataFrame(df_mod["selected_rows"]).shape[0]
        st.write(f"{nb_sel_samples } sample{'s'[:nb_sel_samples^1]} selected")
        st.session_state.force_samp_validation = True

        placeholder_lib_table = st.empty()
        with placeholder_lib_table.container():
            if st.session_state.force_samp_validation:
                with st.form("Library table"):
                    st.markdown("Select libraries to filter")
                    libs = get_libraries(
                        table_name=st.session_state.table_name,
                        libraries=library,
                        samples=pd.DataFrame(df_mod["selected_rows"]),
                        supported_archives=supported_archives,
                    )
                    lib_sel = AgGrid(
                        libs,
                        gridOptions=gridOptions_library,
                        data_return_mode="filtered",
                        update_mode="selection_changed",
                    )
                    try:
                        lib_mod = pd.DataFrame(lib_sel["selected_rows"]).drop(
                            "_selectedRowNodeInfo", axis=1
                        )
                    except KeyError:
                        lib_mod = pd.DataFrame(lib_sel["selected_rows"])

                    if st.form_submit_button(
                        "Validate library selection", type="primary"
                    ):
                        if len(lib_mod) == 0:
                            st.error(
                                "You didn't select any library! Please select at least one library."
                            )
                        else:
                            st.session_state.force_lib_validation = True

        placeholder_buttons = st.empty()

        with placeholder_buttons.container():
            (
                button_sample_table,
                button_libraries,
                button_fastq,
                button_samplesheet_eager,
                button_samplesheet_mag,
                button_samplesheet_taxprofiler,
                button_samplesheet_ameta,
                button_bibtex,
            ) = st.columns(8)

            if (
                st.session_state.force_samp_validation
                and st.session_state.force_lib_validation
            ):
                # Calculate the fastq file size of the selected libraries
                acc_table = prepare_accession_table(
                    pd.DataFrame(df_mod["selected_rows"]),
                    lib_mod,
                    st.session_state.table_name,
                    supported_archives,
                )["df"]
                total_size = (
                    acc_table["download_sizes"]
                    .apply(lambda r: sum([int(s) for s in r.split(";")]))
                    .sum(axis=0)
                )

                if total_size > 1e12:
                    total_size_str = f"{total_size / 1e12:.2f}TB"
                else:
                    total_size_str = f"{total_size / 1e9:.2f}GB"

                ###################
                ## SAMPLE TABLE ##
                ###################

                with button_sample_table:
                    st.download_button(
                        label="Download AncientMetagenomeDir Sample Table",
                        data=(
                            pd.DataFrame(df_mod["selected_rows"])
                            .drop("_selectedRowNodeInfo", axis=1)
                            .to_csv(sep="\t", index=False)
                            .encode("utf-8")
                        ),
                        file_name="AncientMetagenomeDir_filtered_samples.tsv",
                    )

                ###################
                ## LIBRARY TABLE ##
                ###################

                if st.session_state.table_name in ["ancientmetagenome-environmental"]:
                    col_drop = ["archive_accession"]
                else:
                    col_drop = ["archive_accession", "sample_host"]

                with button_libraries:
                    st.download_button(
                        label="Download AncientMetagenomeDir Library Table",
                        data=(
                            lib_mod.drop(col_drop, axis=1).to_csv(sep="\t", index=False)
                        ).encode("utf-8"),
                        file_name="AncientMetagenomeDir_filtered_libraries.tsv",
                    )

                ############################
                ## FASTQ DOWNLOAD SCRIPTS ##
                ############################
                with button_fastq:
                    if st.session_state.dl_method == "nf-core/fetchngs":
                        st.download_button(
                            label=f"Download nf-core/fetchNGS input accession list",
                            help=f"approx. {total_size_str} of sequencing data selected",
                            data=prepare_accession_table(
                                pd.DataFrame(df_mod["selected_rows"]),
                                lib_mod,
                                st.session_state.table_name,
                                supported_archives,
                            )["df"]["archive_accession"]
                            .to_csv(sep="\t", header=False, index=False)
                            .encode("utf-8"),
                            file_name="AncientMetagenomeDir_nf_core_fetchngs_input_table.tsv",
                        )
                    elif st.session_state.dl_method == "aspera":
                        st.download_button(
                            label="Download Aspera sample download script",
                            help=f"approx. {total_size_str} of sequencing data selected",
                            data=prepare_accession_table(
                                pd.DataFrame(df_mod["selected_rows"]),
                                lib_mod,
                                st.session_state.table_name,
                                supported_archives,
                            )["aspera_script"],
                            file_name="AncientMetagenomeDir_aspera_download_script.sh",
                        )
                    elif st.session_state.dl_method == "sratookit":
                        st.download_button(
                            label="Download SRAtoolkit/fasterq-dump sample download script",
                            help=f"approx. {total_size_str} of sequencing data selected",
                            data=prepare_accession_table(
                                pd.DataFrame(df_mod["selected_rows"]),
                                lib_mod,
                                st.session_state.table_name,
                                supported_archives,
                            )["fasterq_dump_script"],
                            file_name="AncientMetagenomeDir_sratoolkit_download_script.sh",
                        )
                    else:
                        st.download_button(
                            label="Download Curl sample download script",
                            help=f"approx. {total_size_str} of sequencing data selected",
                            data=prepare_accession_table(
                                pd.DataFrame(df_mod["selected_rows"]),
                                lib_mod,
                                st.session_state.table_name,
                                supported_archives,
                            )["curl_script"],
                            file_name="AncientMetagenomeDir_curl_download_script.sh",
                        )

                #################
                ## EAGER TABLE ##
                #################
                with button_samplesheet_eager:
                    st.download_button(
                        label="Download nf-core/eager input TSV",
                        data=prepare_eager_table(
                            pd.DataFrame(df_mod["selected_rows"]),
                            lib_mod,
                            st.session_state.table_name,
                            supported_archives,
                        )
                        .to_csv(sep="\t", index=False)
                        .encode("utf-8"),
                        file_name="AncientMetagenomeDir_nf_core_eager_input_table.tsv",
                    )

                    #######################
                    ## NF-CORE/MAG TABLE ##
                    #######################
                with button_samplesheet_mag:
                    mag_table_single, mag_table_paired = prepare_mag_table(
                        pd.DataFrame(df_mod["selected_rows"]),
                        lib_mod,
                        st.session_state.table_name,
                        supported_archives,
                    )
                    zip_file = zipfile.ZipFile(
                        "AncientMetagenomeDir_nf_core_mag_input.zip", mode="w"
                    )
                    if not mag_table_single.empty:
                        mag_table_single.to_csv(
                            "AncientMetagenomeDir_nf_core_mag_input_single_table.csv",
                            index=False,
                        )
                        zip_file.write(
                            "AncientMetagenomeDir_nf_core_mag_input_single_table.csv"
                        )
                        os.remove(
                            "AncientMetagenomeDir_nf_core_mag_input_single_table.csv"
                        )
                    if not mag_table_paired.empty:
                        mag_table_paired.to_csv(
                            "AncientMetagenomeDir_nf_core_mag_input_paired_table.csv",
                            index=False,
                        )
                        zip_file.write(
                            "AncientMetagenomeDir_nf_core_mag_input_paired_table.csv"
                        )
                        os.remove(
                            "AncientMetagenomeDir_nf_core_mag_input_paired_table.csv"
                        )
                    zip_file.close()
                    with open(
                        "AncientMetagenomeDir_nf_core_mag_input.zip", "rb"
                    ) as zip_file:
                        st.download_button(
                            label="Download nf-core/mag input CSV",
                            data=zip_file,
                            file_name="AncientMetagenomeDir_nf_core_mag_input.zip",
                            mime="application/zip",
                        )
                    os.remove("AncientMetagenomeDir_nf_core_mag_input.zip")

                #######################
                ## TAXPROFILER TABLE ##
                #######################
                with button_samplesheet_taxprofiler:
                    st.download_button(
                        label="Download nf-core/taxprofiler input CSV",
                        data=prepare_taxprofiler_table(
                            pd.DataFrame(df_mod["selected_rows"]),
                            lib_mod,
                            st.session_state.table_name,
                            supported_archives,
                        )
                        .to_csv(index=False)
                        .encode("utf-8"),
                        file_name="AncientMetagenomeDir_nf_core_taxprofiler_input_table.csv",
                    )

                #################
                ## AMETA TABLE ##
                #################
                with button_samplesheet_ameta:
                    st.download_button(
                        label="Download aMeta input TSV",
                        data=prepare_aMeta_table(
                            pd.DataFrame(df_mod["selected_rows"]),
                            lib_mod,
                            st.session_state.table_name,
                            supported_archives,
                        )
                        .to_csv(sep="\t", index=False)
                        .encode("utf-8"),
                        file_name="AncientMetagenomeDir_aMeta_input_table.tsv",
                    )

                #################
                ## BIBTEX FILE ##
                #################
                with button_bibtex:
                    st.download_button(
                        label="Download Citations as BibTex",
                        data=prepare_bibtex_file(pd.DataFrame(df_mod["selected_rows"])),
                        file_name="AncientMetagenomeDir_bibliography.bib",
                    )

                st.markdown(
                    "ℹ️ _By default all download scripts/inputs include ALL libraries of the selected samples. \n Review the AncientMetagenomeDir library table prior using any other table, to ensure usage of relevant libraries!_"
                )
                st.markdown(
                    "⚠️ _We provide no warranty to the accuracy of the generated input sheets._"
                )

                if st.button("Start New Selection", type="primary"):
                    st.session_state.compute = False
                    st.session_state.table_name = "No table selected"
                    st.session_state.force_samp_validation = False
                    st.session_state.force_lib_validation = False
                    placeholder_buttons.empty()
                    placeholder_lib_table.empty()



from amdirt.download import download


def test_download():
    table = "ancientmetagenome-hostassociated"
    table_type = "samples"
    release = "v23.12.0"

    d = download(table, table_type, release, output=".")
    assert d == "ancientmetagenome-hostassociated_samples_v23.12.0.tsv"



from amdirt.validate.domain import DatasetValidator
from amdirt.core import get_json_path, logger
from amdirt.core.ena import ENAPortalAPI
from amdirt.validate.exceptions import NetworkError
import json

import sys
import pandas as pd

def run_autofill(accession, table_name=None, schema=None, dataset=None, sample_output=None, library_output=None, verbose=False):
    """Autofill the metadata of a table from ENA

    Args:
        accession (tuple(str)): ENA project accession. Multiple accessions can be space separated (e.g. PRJNA123 PRJNA456)
        table_name (str): Name of the table to be filled
        schema (str): Path to the schema file
        dataset (str): Path to the dataset file
        sample_output (str): Path to the sample output table file
        library_output (str): Path to the library output table file

    Returns:
        pd.DataFrame: ENA metadata run level table
    """

    if schema is None and dataset is None:
        json_conf = get_json_path()
        with open(json_conf) as c:
            tables = json.load(c)
            samples = tables["samples"]
            samples_schema = tables["samples_schema"]
            libraries = tables["libraries"]
            libraries_schema = tables["libraries_schema"]
        if table_name not in samples:
            raise Exception("Table name not found in AncientMetagenomeDir file")
    else:
        logger.error("Not implemented yet")

    sample = DatasetValidator(
        schema=samples_schema[table_name], dataset=samples[table_name]
    )
    libraries = DatasetValidator(
        schema=libraries_schema[table_name], dataset=libraries[table_name]
    )

    libraries.to_rich()

    sample_df = sample.dataset.iloc[:0, :].copy()
    libraries_df = libraries.dataset.iloc[:0, :].copy()

    ena = ENAPortalAPI()
    if ena.status():
        logger.info("ENA API is up")
    else:
        raise NetworkError("ENA API is unreachable")
    
    query_dict = list()
    for a in accession:
        query_res = ena.query(a, fields=[
            "study_accession",
            "run_accession",
            "secondary_sample_accession",
            "sample_alias",
            "fastq_ftp",
            "fastq_md5",
            "fastq_bytes",
            "library_name",
            "instrument_model",
            "library_layout",
            "library_strategy",
            "read_count",
        ])
        query_dict += query_res
    df_out = pd.DataFrame.from_dict(query_dict)

    df_out.rename(
        columns={
            "study_accession": "archive_project",
            "run_accession":"archive_data_accession",
            "secondary_sample_accession": "archive_sample_accession",
            "sample_alias": "sample_name",
            "fastq_ftp": "download_links",
            "fastq_md5": "download_md5s",
            "fastq_bytes": "download_sizes"
        },
        inplace=True
        )
    
    # sample table
    sample_out = df_out.copy(deep=True)
    sample_out.rename(
        columns={
            "archive_sample_accession": "archive_accession",
        },
        inplace=True
    )
    for col in sample_df.columns:
        if col not in sample_out.columns:
            sample_out[col] = None
    sample_out = sample_out[sample_df.columns]
    sample_out = sample_out.loc[:,~sample_out.columns.duplicated()].copy()
    sample_out = sample_out.drop_duplicates(subset=["archive_accession"])
    sample_out = sample_out.astype(sample_df.dtypes.to_dict())

    # library table
    lib_out = df_out.copy(deep=True)
    for col in libraries_df.columns:
        if col not in lib_out.columns:
            lib_out[col] = None
    lib_out = lib_out[libraries_df.columns]
    lib_out = lib_out.loc[:,~lib_out.columns.duplicated()].copy()
    lib_out['read_count'] = lib_out['read_count'].str.replace(",", "",
                                                              regex=False)
    lib_out = lib_out.astype(libraries_df.dtypes.to_dict())

    if library_output:
        logger.info(f"Found {lib_out.shape[0]} libraries")
        logger.info(f"Writing libraries metadata to {library_output}")
        lib_out.to_csv(library_output, index=False, sep="\t")
    if sample_output:
        logger.info(f"Found {sample_out.shape[0]} samples")
        logger.info(f"Writing samples metadata to {sample_output}")
        logger.warning("Sample name must match that reported in publication and/or sample-level table. ENA reported sample-name may not be correct! Check before submission.")
        sample_out.to_csv(sample_output, index=False, sep="\t")



from amdirt.validate.application import AMDirValidator
from amdirt.validate.exceptions import DatasetValidationError
import warnings
import pandas as pd
from amdirt.core import logger, get_remote_resources
from os.path import join


def merge_new_df(
    dataset,
    table_type,
    table_name,
    markdown,
    outdir,
    verbose,
    schema_check=True,
    line_dup=True,
    columns=True,
):
    """Merge a new dataset with the remote master dataset

    Args:
        dataset (Path): Path to new dataset
        table_type (str): Type of table to merge (samples or libraries)
        table_name (str): Kind of table to merge (e.g. ancientmetagenome-hostassociated, ancientmetagenome-environmental, etc.)
        markdown (bool): Log in markdown format
        outdir (Path): Path to output directory
        verbose (bool): Enable verbose mode
        schema_check (bool, optional): Enable schema check. Defaults to True.
        line_dup (bool, optional): Enable line duplication check. Defaults to True.
        columns (bool, optional): Enable columns presence/absence check. Defaults to True.

    Raises:
        ValueError: Table type must be either 'samples' or 'libraries'
        ValueError: Table name not found in AncientMetagenomeDir file
        DatasetValidationError: New dataset is not valid
    """
    remote_resources = get_remote_resources()

    if table_type not in ["samples", "libraries"]:
        raise ValueError("table_type must be either 'samples' or 'libraries'")
    if table_name not in remote_resources[table_type]:
        raise ValueError("table_name not found in AncientMetagenomeDir file")
    if not verbose:
        warnings.filterwarnings("ignore")

    schema = remote_resources[f"{table_type}_schema"][table_name]
    dataset_valid = list()
    v = AMDirValidator(schema, dataset)
    dataset_valid.append(v.parsing_ok)
    if schema_check and v.parsing_ok:
        dataset_valid.append(v.validate_schema())
    if line_dup and v.parsing_ok:
        dataset_valid.append(v.check_duplicate_rows())
    if columns and v.parsing_ok:
        dataset_valid.append(v.check_columns())

    dataset_valid = all(dataset_valid)
    if dataset_valid is False:
        if markdown:
            v.to_markdown()
        else:
            v.to_rich()
        raise DatasetValidationError("New Dataset is not valid")

    else:
        remote_dataset = pd.read_table(
            remote_resources[table_type][table_name], dtype=dict(v.dataset.dtypes)
        )

        logger.info("New Dataset is valid")
        logger.info(
            f"Merging new dataset with remote {table_name} {table_type} dataset"
        )
        dataset = pd.concat([remote_dataset, v.dataset])
        dataset.drop_duplicates(inplace=True)
        dataset.to_csv(
            join(outdir, f"{table_name}_{table_type}.tsv"),
            sep="\t",
            na_rep="NA",
            index=False,
        )
        logger.info(
            f"New {table_name} {table_type} dataset written to {join(outdir,f'{table_name}_{table_type}.tsv')}"
        )



from amdirt.core import (
    logger,
    get_amdir_tags,
    get_remote_resources,
    check_allowed_values,
)
import requests


def download(table: str, table_type: str, release: str, output: str = ".") -> str:
    """
    Download a table from the amdirt repository.

    Parameters
    ----------
    table : str
        The AncientMetagenomeDir table to download.
    table_type : str
        The type of table to download. Allowed values are ['samples', 'libraries'].
    release : str
        The release of the table to download. Must be a valid release tag.
    output: str
        The output directory to save the table. Default is the current directory.

    Returns
    -------
    str:
        The path to the downloaded table.

    Raises
    ------
    ValueError
        If an invalid table is provided.
    ValueError
        If an invalid table type is provided.
    ValueError
        If an invalid release is provided.
    """

    resources = get_remote_resources()
    tags = get_amdir_tags()
    if tags != ["master"]:
        if check_allowed_values(tags, release) is False:
            raise ValueError(f"Invalid release: {release}. Allowed values are {tags}")

    tables = resources["samples"]
    if check_allowed_values(tables, table) is False:
        raise ValueError(f"Invalid table: {table}. Allowed values are {tables}")

    if check_allowed_values(["samples", "libraries"], table_type) is False:
        raise ValueError(
            f"Invalid table type: {table_type}. Allowed values are ['samples', 'libraries']"
        )
    table_filename = f"{table}_{table_type}_{release}.tsv"
    logger.info(
        f"Downloading {table} {table_type} table from {release} release, saving to {output}/{table_filename}"
    )
    t = requests.get(resources[table_type][table].replace("master", release))
    with open(table_filename, "w") as fh:
        fh.write(t.text)

    return table_filename



import click
from amdirt import __version__

from amdirt.validate import run_validation
from amdirt.viewer import run_app
from amdirt.convert import run_convert
from amdirt.core import get_json_path, get_amdir_tags, get_latest_tag
from amdirt.autofill import run_autofill
from amdirt.merge import merge_new_df
from amdirt.download import download as download_amdir
from json import load


class MutuallyExclusiveOption(click.Option):
    # Credits goes to Stan Chang for this code snippet
    # https://gist.github.com/stanchan/bce1c2d030c76fe9223b5ff6ad0f03db

    def __init__(self, *args, **kwargs):
        self.mutually_exclusive = set(kwargs.pop("mutually_exclusive", []))
        help = kwargs.get("help", "")
        if self.mutually_exclusive:
            ex_str = ", ".join(self.mutually_exclusive)
            kwargs["help"] = help + (
                " NOTE: This argument is mutually exclusive with "
                " arguments: [" + ex_str + "]."
            )
        super(MutuallyExclusiveOption, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        if self.mutually_exclusive.intersection(opts) and self.name in opts:
            raise click.UsageError(
                "Illegal usage: `{}` is mutually exclusive with "
                "arguments `{}`.".format(self.name, ", ".join(self.mutually_exclusive))
            )

        return super(MutuallyExclusiveOption, self).handle_parse_result(ctx, opts, args)


def get_table_list():
    json_path = get_json_path()
    with open(json_path, "r") as f:
        table_list = load(f)
        return list(table_list["samples"].keys())


@click.group()
@click.version_option(__version__)
@click.pass_context
@click.option("--verbose", is_flag=True, help="Verbose mode")
def cli(ctx, verbose, no_args_is_help=True, **kwargs):
    """\b
    amdirt: Performs validity check of AncientMetagenomeDir datasets
    Authors: amdirt development team and the SPAAM community
    Homepage & Documentation: https://github.com/SPAAM-community/amdirt
    \b
    """
    ctx.ensure_object(dict)
    ctx.obj["verbose"] = verbose
    pass


####################
# Validation  tool #
####################


@cli.command()
@click.argument("dataset", type=click.Path(exists=True))
@click.argument("schema", type=click.Path(exists=True))
@click.option("-s", "--schema_check", is_flag=True, help="Turn on schema checking.")
@click.option(
    "-d", "--line_dup", is_flag=True, help="Turn on line duplicate line checking."
)
@click.option(
    "-c", "--columns", is_flag=True, help="Turn on column presence/absence checking."
)
@click.option("-i", "--doi", is_flag=True, help="Turn on DOI duplicate checking.")
@click.option(
    "--multi_values",
    multiple=True,
    help="Check multi-values column for duplicate values.",
)
@click.option(
    "-a",
    "--online_archive",
    is_flag=True,
    help="Turn on ENA accession validation",
)
@click.option(
    "--remote",
    type=click.Path(),
    default=None,
    help="[Optional] Path/URL to remote reference sample table for archive accession validation",
)
@click.option("-m", "--markdown", is_flag=True, help="Output is in markdown format")
@click.pass_context
def validate(ctx, no_args_is_help=True, **kwargs):
    """\b
    Run validity check of AncientMetagenomeDir datasets
    \b
    DATASET: path to tsv file of dataset to check
    SCHEMA: path to JSON schema file
    """
    run_validation(**kwargs, **ctx.obj)


###############################
# Interactive viewing/filtering  tool #
###############################


@cli.command()
@click.option(
    "-t",
    "--tables",
    type=click.Path(exists=True),
    help="JSON file listing AncientMetagenomeDir tables",
)
@click.pass_context
def viewer(ctx, no_args_is_help=True, **kwargs):
    """Launch interactive filtering tool"""
    run_app(**kwargs, **ctx.obj)


###################
# Conversion tool #
###################


@cli.command()
@click.argument("samples", type=click.Path(exists=True))
@click.argument("table_name", type=str)
@click.option(
    "-t",
    "--tables",
    type=click.Path(exists=True),
    help="(Optional) JSON file listing AncientMetagenomeDir tables",
)
@click.option(
    "--libraries",
    type=click.Path(readable=True, file_okay=True, dir_okay=False, exists=True),
    help=("(Optional) Path to a pre-filtered libraries table"),
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["librarymetadata"],
)
@click.option(
    "--librarymetadata",
    is_flag=True,
    help="Generate AncientMetagenomeDir libraries table of all samples in input table",
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["libraries"],
)
@click.option(
    "-o",
    "--output",
    type=click.Path(writable=True, dir_okay=True, file_okay=False),
    default=".",
    show_default=True,
    help="conversion output directory",
)
@click.option(
    "--bibliography",
    is_flag=True,
    help="Generate BibTeX file of all publications in input table",
)
@click.option(
    "--curl",
    is_flag=True,
    help="Generate bash script with curl-based download commands for all libraries of samples in input table",
)
@click.option(
    "--aspera",
    is_flag=True,
    help="Generate bash script with Aspera-based download commands for all libraries of samples in input table",
)
@click.option(
    "--fetchngs",
    is_flag=True,
    help="Convert filtered samples and libraries tables to nf-core/fetchngs input tables",
)
@click.option(
    "--sratoolkit",
    is_flag=True,
    help="Generate bash script with SRA Toolkit fasterq-dump based download commands for all libraries of samples in input table",
)
@click.option(
    "--eager",
    is_flag=True,
    help="Convert filtered samples and libraries tables to eager input tables",
)
@click.option(
    "--ameta",
    is_flag=True,
    help="Convert filtered samples and libraries tables to aMeta input tables",
)
@click.option(
    "--mag",
    is_flag=True,
    help="Convert filtered samples and libraries tables to nf-core/mag input tables",
)
@click.option(
    "--taxprofiler",
    is_flag=True,
    help="Convert filtered samples and libraries tables to nf-core/taxprofiler input tables",
)
@click.pass_context
def convert(ctx, no_args_is_help=True, **kwargs):
    """\b
    Converts filtered samples and libraries tables to eager, ameta, taxprofiler, and fetchNGS input tables

    Note: When supplying a pre-filtered libraries table with `--libraries`, the corresponding sample table is still required!

    \b
    SAMPLES: path to filtered AncientMetagenomeDir samples tsv file
    TABLE_NAME: name of table to convert
    """
    run_convert(**kwargs, **ctx.obj)


#################
# Autofill tool #
#################


@cli.command()
@click.argument("accession", type=str, nargs=-1)
@click.option(
    "-n",
    "--table_name",
    type=click.Choice(get_table_list()),
    default="ancientmetagenome-hostassociated",
    show_default=True,
)
@click.option(
    "-l",
    "--library_output",
    type=click.Path(writable=True),
    help="path to library output table file",
)
@click.option(
    "-s",
    "--sample_output",
    type=click.Path(writable=True),
    help="path to sample output table file",
)
@click.pass_context
def autofill(ctx, no_args_is_help=True, **kwargs):
    """\b
    Autofills library and/or sample table(s) using ENA API and accession numbers
    \b

    ACCESSION: ENA accession(s). Multiple accessions can be space separated (e.g. PRJNA123 PRJNA456)
    """
    run_autofill(**kwargs, **ctx.obj)


################
# Merging tool #
################


@cli.command()
@click.argument("dataset", type=click.Path(exists=True))
@click.option(
    "-n",
    "--table_name",
    type=click.Choice(get_table_list()),
    default="ancientmetagenome-hostassociated",
    show_default=True,
)
@click.option(
    "-t",
    "--table_type",
    type=click.Choice(["samples", "libraries"]),
    default="libraries",
    show_default=True,
)
@click.option("-m", "--markdown", is_flag=True, help="Output is in markdown format")
@click.option(
    "-o",
    "--outdir",
    type=click.Path(writable=True),
    default=".",
    show_default=True,
    help="path to sample output table file",
)
@click.pass_context
def merge(ctx, no_args_is_help=True, **kwargs):
    """\b
    Merges new dataset with existing table
    \b

    DATASET: path to tsv file of new dataset to merge
    """
    merge_new_df(**kwargs, **ctx.obj)


@cli.command()
@click.option(
    "-t",
    "--table",
    help="AncientMetagenomeDir table to download",
    type=click.Choice(get_table_list()),
    default="ancientmetagenome-hostassociated",
    show_default=True,
)
@click.option(
    "-y",
    "--table_type",
    help="Type of table to download",
    type=click.Choice(["samples", "libraries"]),
    default="samples",
    show_default=True,
)
@click.option(
    "-r",
    "--release",
    help="Release tag to download",
    type=click.Choice(get_amdir_tags()),
    default=get_latest_tag(get_amdir_tags()),
    show_default=True,
)
@click.option(
    "-o",
    "--output",
    help="Output directory",
    type=click.Path(writable=True),
    default=".",
    show_default=True,
)
def download(no_args_is_help=True, **kwargs):
    """\b
    Download a table from the amdirt repository
    """
    download_amdir(**kwargs)


if __name__ == "__main__":
    cli()



class Error(Exception):
    """Base class for exceptions in this module."""

    pass


class DatasetValidationError(Error):
    """Exception raised for errors in validating the dataset againsy the standards.

    Attributes:
        message -- explanation of the error
    """

    name = "Dataset Validation Error"

    def __init__(self, message):
        self.message = message


class DuplicateError(Error):
    """Exception raised for errors due to duplicated rows"

    Attributes:
        message -- explanation of the error
    """

    name = "Duplicated Row Error"

    def __init__(self, message):
        self.message = message


class DOIDuplicateError(Error):
    """Exception raised for errors due to duplicates in standards

    Attributes:
        message -- explanation of the error
    """

    name = "Duplicated DOI Error"

    def __init__(self, message):
        self.message = message


class ColumnDifferenceError(Error):
    """Exception raised for errors due to missing/added columns

    Attributes:
        message -- explanation of the error
    """

    name = "Different Columns Error"

    def __init__(self, message):
        self.message = message


class ParsingError(Error):
    """Exception raised for parsing errors

    Attributes:
        message -- explanation of the error
    """

    name = "Parsing Error"

    def __init__(self, message):
        self.message = message

class NetworkError(Error):
    """Exception raised for parsing errors

    Attributes:
        message -- explanation of the error
    """

    name = "Network Error"

    def __init__(self, message):
        self.message = message


class DuplicateEntryError(Error):
    """Exception raised for errors due to duplicated entries in columns"

    Attributes:
        message -- explanation of the error
    """

    name = "Duplicated Entries in column Error"

    def __init__(self, message):
        self.message = message



import json
import pandas as pd
from amdirt.validate import exceptions
from amdirt.core import logger
from io import StringIO
from pathlib import Path
from rich.table import Table
from rich.console import Console
from dataclasses import dataclass
from jsonschema import Draft7Validator, exceptions as json_exceptions
from typing import AnyStr, BinaryIO, TextIO, Union
import requests
import re

Schema = Union[AnyStr, BinaryIO, TextIO]
Dataset = Union[AnyStr, BinaryIO, TextIO]


@dataclass
class DFError:
    """Class to store DataFrame validation errors"""

    error: str
    source: str
    column: str
    row: str
    message: str
    line_offset: int = (
        2  # Offset to add to row number to account for header and 0 based indexing
    )

    def to_dict(self):
        try:
            column = str(self.column)
            row = str(int(self.row) + self.line_offset)
        except (ValueError, TypeError):
            column = "-"
            row="-"
        return {
            "Error": str(self.error),
            "Source": str(self.source),
            "Column": column,
            "Row": row,
            "Message": str(self.message),
        }

    def to_list(self):
        try:
            column = str(self.column)
            row = str(int(self.row) + self.line_offset)
        except (ValueError, TypeError):
            column = "-"
            row="-"
        return [
        str(i)
        for i in [self.error, self.source, column, row , self.message]
        ]

class DatasetValidator:
    """Dataset as DataFrame validation class"""

    def __init__(self, schema: Schema, dataset: Dataset):
        """Dataset validation class

        Attributes:
            errors (list): List of DFError objects
            dataset_name (str): Dataset name
            schema_name (str): Schema name
            schema (dict): JSON schema
            dataset (pd.DataFrame): Dataset as pandas dataframe
            dataset_json (dict): Dataset as dictionary

        Args:
            schema (Schema): Path to schema in json format
            dataset (Dataset): Path to dataset in tsv format

        """
        self.errors = list()
        self.dataset_path = Path(dataset)
        if str(schema).startswith("http"):
            self.schema_path = schema
        else:
            self.schema_path = Path(schema)
        self.dataset_name = Path(dataset).name
        self.schema_name = Path(schema).name
        self.schema = self.read_schema(schema)
        if self.schema:
            self.dataset = self.read_dataset(dataset, self.schema)
        else:
            self.dataset = False
        self.parsing_ok = (
            True if isinstance(self.dataset, pd.DataFrame) and self.schema else False
        )
        if self.parsing_ok:
            self.dataset_json = self.dataset_to_json()

    def __repr__(self):
        return (
            f"DatasetValidator(schema={self.schema_name}, dataset={self.dataset_name})"
        )

    def add_error(self, err: DFError):
        self.errors.append(err)

    def read_schema(self, schema: Schema) -> dict:
        """Read JSON schema from file or string

        Args:
            schema (str): Path to schema in json format
        Returns:
            dict: JSON schema
        """
        try:
            if str(schema).startswith("http"):
                res = requests.get(schema)
                if res.status_code == 200:
                    return res.json()
                else:
                    raise Exception("Could not fetch schema from URL")
            else:
                with open(schema, "r") as s:
                    return json.load(s)
        except json.JSONDecodeError as e:
            msg = str(e.with_traceback(e.__traceback__))
            self.add_error(
                DFError(
                    error="Schema Error",
                    source=e.msg,
                    column=str(*re.findall(".*column.(\d+).*", msg)),
                    row=int(re.findall(".*line.(\d+).*", msg)[0]),
                    message="JSON parsing error",
                )
            )
            return False

    def read_dataset(self, dataset: Dataset, schema: dict) -> pd.DataFrame:
        """ "Read dataset from file or string
        Args:
            dataset (str): Path to dataset in tsv format
            schema (dict): Parsed schema as dictionary (from read_schema)
        Returns:
            pd.DataFrame: Dataset as pandas dataframe
        """
        string_to_dtype_conversions = {
            "string": str,
            "integer": pd.Int64Dtype(),
            "number": float,
        }
        column_dtypes = {}
        for column_keys in schema["items"]["properties"]:
            coltype = schema["items"]["properties"][column_keys]["type"]
            if isinstance(coltype, list):
                if ("null" in coltype and len(coltype) > 2) or (
                    len(coltype) > 1 and "null" not in coltype
                ):
                    self.add_error(
                        DFError(
                            error="Schema Error",
                            source=coltype,
                            column=column_keys,
                            row="",
                            message="No mixed data types allowed",
                        )
                    )
                    return False
                else:
                    coltype = coltype[0]
            if coltype in string_to_dtype_conversions:
                column_dtypes[column_keys] = string_to_dtype_conversions[coltype]
            elif coltype == "null":
                self.add_error(
                    DFError(
                        error="Schema Error",
                        source=coltype,
                        column=column_keys,
                        row="",
                        message="Default/first type of column in schema can not be null",
                    )
                )
                return False
            else:
                column_dtypes[column_keys] = coltype
        try:
            return pd.read_table(dataset, sep="\t", dtype=column_dtypes)
        except (AttributeError, pd.errors.ParserError, ValueError) as e:
            self.add_error(
                DFError(
                    "Dataset Parsing Error",
                    self.dataset_name,
                    '-',
                    '-',
                    e,
                )
            )
            return False

    def check_columns(self) -> bool:
        """Checks if dataset has all required columns

        Returns:
            bool: True if dataset has all required columns, False otherwise
        """
        col_diff = set(self.dataset.columns).difference(
            set(self.schema["items"]["required"])
        )
        try:
            if len(col_diff) != 0:
                raise exceptions.ColumnDifferenceError(
                    f"Dataset has different columns compared to schema {col_diff}"
                )
            else:
                return True
        except exceptions.ColumnDifferenceError as e:
            for c in col_diff:
                self.add_error(DFError(e.name, c, None, None, e.message))
            return False

    def dataset_to_json(self) -> dict:
        """Convert dataset from Pandas DataFrame to JSON

        Returns:
            dict: Dataset as dictionary
        """
        return json.load((StringIO(self.dataset.to_json(orient="records"))))

    def validate_schema(self) -> bool:
        """Validate dataset against JSON schema

        Returns:
            bool: True if dataset is valid, False otherwise
        """
        validator = Draft7Validator(self.schema)
        err_cnt = 0
        for err in validator.iter_errors(self.dataset_json):
            self.add_error(self.cleanup_errors(err))
            err_cnt += 1
        if err_cnt > 0:
            return False
        else:
            return True

    def cleanup_errors(self, error: json_exceptions.ValidationError) -> DFError:
        """Cleans up JSON schema validation errors

        Args:
            error (json_exceptions.ValidationError): JSON schema validation error
        Returns:
            DFError: Cleaned DataFrame error
        """
        err_column = list(error.path)[-1]
        if "enum" in error.schema:
            if len(error.schema["enum"]) > 3:
                error.message = f"'{error.instance}' is not an accepted value.\nPlease check [link={self.schema['items']['properties'][err_column]['$ref']}]{self.schema['items']['properties'][err_column]['$ref']}[/link]"
        err_line = str(error.path[0])
        return DFError(
            "Schema Validation Error",
            error.instance,
            err_column,
            err_line,
            error.message,
        )

    def check_duplicate_rows(self) -> bool:
        """Checks for duplicated rows in dataset

        Returns:
            bool: True if dataset has no duplicated rows, False otherwise
        """
        try:
            dup_df = self.dataset[self.dataset.duplicated(keep=False)]
            if len(dup_df) > 0:
                dup_rows = (
                    dup_df.groupby(dup_df.columns.tolist(), dropna=False)
                    .apply(lambda x: tuple(x.index))
                    .tolist()
                )
                raise exceptions.DuplicateError(f"Rows \n{dup_rows} are duplicated")
        except exceptions.DuplicateError as e:
            for r in dup_rows:
                self.add_error(DFError(e.name, r, None, r, f"Rows {r} are duplicated"))
            return False
        return True

    def to_rich(self):
        """Generate rich output table for console display

        Returns:
            bool: True if dataset is valid
        Raises:
            SystemExit: If dataset is invalid
        """

        table = Table(
            title=f"amdirt Validation Report of {self.dataset_name} against {self.schema_name}"
        )
        columns = ["Error", "Source", "Column", "Row", "Message"]
        for column in columns:
            table.add_column(column, style="red", overflow="fold")
        for error in self.errors:
            table.add_row(*(error.to_list()))

        console = Console()
        if len(self.errors) > 0:
            console.print(table)
            raise SystemExit(f"Invalid dataset {self.dataset_name}")
        else:
            logger.info(f"{self.dataset_name} is valid")
            return True

    def to_markdown(self) -> bool:
        """Generate markdown output table for github display

        Returns:
            bool: True if dataset is valid
        Raises:
            SystemExit: If dataset is invalid
        """
        if len(self.errors) > 0:
            df = pd.concat([pd.Series(error.to_dict()).to_frame().transpose() for error in self.errors])
            raise SystemExit(
                f"Invalid dataset `{self.dataset_name}`\n\n{df.to_markdown(index=False)}"
            )
        else:
            print(f"`{self.dataset_name}` is valid")
            return True



# from amdirt import logger
from distutils.log import warn
import sys
from streamlit.web import cli as stcli
from pathlib import Path
from amdirt.core import get_json_path, logger
import warnings


def run_app(tables=None, verbose=False):
    """
    Run the amdirt interactive filtering application

    Args:
        tables (str): path to JSON file listing AncientMetagenomeDir tables
    """
    if not warnings:
        warnings.filterwarnings("ignore")
    directory = Path(__file__).parent.resolve()
    app = "streamlit.py"
    if tables is None:
        config_path = get_json_path()
    else:
        config_path = tables
    app_path = f"{directory}/{app}"

    sys.argv = [
        "streamlit", 
        "run", 
        app_path, 
        "--", 
        "--config", 
        config_path
    ]
    logger.info("\n[amdirt] To close app, press on your keyboard: ctrl+c\n")
    sys.exit(stcli.main())



import os
from amdirt.validate import AMDirValidator


def test_project_multiple_doi(test_data_dir):
    valid_dataset = os.path.join(test_data_dir, "valid.tsv")
    invalid_doi = os.path.join(test_data_dir, "invalid_project_multiple_DOI.tsv")
    schema = os.path.join(test_data_dir, "schema.json")

    valid = AMDirValidator(schema=schema, dataset=valid_dataset)
    invalid = AMDirValidator(schema=schema, dataset=invalid_doi)

    assert valid.check_duplicate_dois() is True

    assert invalid.check_duplicate_dois() is False


def test_multi_values(test_data_dir):
    valid_dataset = os.path.join(test_data_dir, "valid.tsv")
    invalid_multiple_values = os.path.join(
        test_data_dir, "invalid_multi_archive_accession.tsv"
    )
    schema = os.path.join(test_data_dir, "schema.json")

    valid = AMDirValidator(schema=schema, dataset=valid_dataset)
    invalid = AMDirValidator(schema=schema, dataset=invalid_multiple_values)

    assert valid.check_multi_values() is True
    assert invalid.check_multi_values() is False


def test_check_sample_accession(test_data_dir):
    valid_dataset = os.path.join(test_data_dir, "valid_newline.tsv")
    invalid_dataset = os.path.join(
        test_data_dir, "invalid_sample_accession_newline.tsv"
    )
    remote_dataset = os.path.join(test_data_dir, "valid.tsv")
    schema = os.path.join(test_data_dir, "schema.json")

    valid = AMDirValidator(schema=schema, dataset=valid_dataset)
    invalid = AMDirValidator(schema=schema, dataset=invalid_dataset)

    assert valid.check_sample_accession(remote=remote_dataset) is True
    assert invalid.check_sample_accession(remote=remote_dataset) is False



from amdirt.core import (
    get_colour_chemistry,
    doi2bib,
    ena,
)


def test_get_colour_chemistry():

    assert get_colour_chemistry("hiseq") == 4
    assert get_colour_chemistry("novaseq") == 2


def test_doi2bib():
    assert doi2bib("10.1038/nature14236")[:18] == " @article{Mnih_201"


def test_ena_portal_status():
    ena_portal = ena.ENAPortalAPI()
    assert ena_portal.status() is True


def test_ena_browser_status():
    ena_browser = ena.ENABrowserAPI()
    assert ena_browser.status() is True



import itertools
from typing import AnyStr, Tuple, Dict
from pandas import DataFrame


def get_sample_diff(local: DataFrame, remote: DataFrame, schema: Dict) -> Tuple[AnyStr]:
    """Returns a diff of the two files.

    Args:
        local(DataFrame): Local sample dataframe.
        remote(DataFrame): Remote sample dataframe.
        schema(dict): sample JSON Schema.

    Returns:
        tuple: A set of the sample only present locally.
    """

    remote_samples = set(
        itertools.chain.from_iterable(
            [i.split(",") for i in remote["archive_accession"]]
        )
    )

    local_samples = set(
        itertools.chain.from_iterable(
            [i.split(",") for i in local["archive_accession"]]
        )
    )

    return tuple(local_samples.difference(remote_samples))



from typing import Tuple, Iterable
from pathlib import Path
import requests
from numpy import where
import pandas as pd
import streamlit as st
from packaging import version
from packaging.version import InvalidVersion
from importlib.resources import files as get_module_dir
import os
import logging
import colorlog
from json import load


pd.options.mode.chained_assignment = None

logging.basicConfig(level=logging.INFO)

handler = colorlog.StreamHandler()
handler.setFormatter(
    colorlog.ColoredFormatter("%(log_color)s%(name)s [%(levelname)s]: %(message)s")
)

logger = colorlog.getLogger("amdirt")
logger.addHandler(handler)
logger.propagate = False


def monkeypatch_get_storage_manager():
    if st.runtime.exists():
        return st.runtime.get_instance().cache_storage_manager
    else:
        # When running in "raw mode", we can't access the CacheStorageManager,
        # so we're falling back to InMemoryCache.
        # https://github.com/streamlit/streamlit/issues/6620
        # _LOGGER.warning("No runtime found, using MemoryCacheStorageManager")
        return (
            st.runtime.caching.storage.dummy_cache_storage.MemoryCacheStorageManager()
        )


st.runtime.caching._data_caches.get_storage_manager = monkeypatch_get_storage_manager


def get_json_path():
    path = get_module_dir("amdirt.assets").joinpath("tables.json")
    return path


def get_remote_resources():
    json_path = get_json_path()
    with open(json_path, "r") as f:
        return load(f)


@st.cache_data
def get_amdir_tags():
    r = requests.get(
        "https://api.github.com/repos/SPAAM-community/AncientMetagenomeDir/tags"
    )
    if r.status_code == 200:
        tags = [
            tag['name']
            for tag in r.json()
            if tag['name'] != 'latest'
        ]
        return [
            tag
            for tag in tags
            if version.parse(tag) >= version.parse("v22.09")
        ]
    else:
        logger.warning(
            "Could not fetch tags from AncientMetagenomeDir. Defaulting to master. Metadata may not yet be officially released."
        )
        return ["master"]


@st.cache_data
def get_latest_tag(tags):
    try:
        return sorted(tags, key=lambda x: version.Version(x))[-1]
    except InvalidVersion:
        if "master" in tags:
            return "master"
        else:
            raise InvalidVersion("No valid tags found")


def check_allowed_values(ref: list, test: str):
    """
    Check if test is in ref
    Args:
        ref(list): List of allowed values
        test(str): value to check
    """

    if test in ref:
        return True
    return False


def get_colour_chemistry(instrument: str) -> int:
    """
    Get number of colours used in sequencing chemistry
    Args:
        instrument(str): Name of the instrument
    Returns:
        int: number of colours used in sequencing chemistry
    """
    chemistry_colours = {
        "bgiseq": 4,
        "miseq": 4,
        "hiseq": 4,
        "genome analyzer": 4,
        "nextseq": 2,
        "novaseq": 2,
    }

    for k in chemistry_colours:
        if k in instrument.lower():
            return chemistry_colours[k]


def doi2bib(doi: str) -> str:
    """
    Return a bibTeX string of metadata for a given DOI.
    """

    url = "http://doi.org/" + doi

    headers = {"accept": "application/x-bibtex"}
    r = requests.get(url, headers=headers)

    return r.text


@st.cache_data
def get_libraries(
    table_name: str,
    samples: pd.DataFrame,
    libraries: pd.DataFrame,
    supported_archives: Iterable[str],
):
    """Get filtered libraries from samples and libraries tables

    Args:
        table_name (str): Name of the table of the table to convert
        samples (pd.DataFrame): Sample table
        libraries (pd.DataFrame): Library table
        supported_archives (Iterable[str]): Supported archives list

    Returns:
        pd.DataFrame: filtered libraries table
    """
    stacked_samples = (
        samples.query("archive in @supported_archives")
        .loc[:, "archive_accession"]
        .str.split(",", expand=True)
        .stack()
        .reset_index(level=0)
        .set_index("level_0")
        .rename(columns={0: "archive_accession"})
        .join(samples.drop("archive_accession", axis=1))
    )

    if table_name in [
        "ancientmetagenome-environmental",
    ]:
        sel_col = ["archive_accession"]
    else:
        sel_col = ["archive_accession", "sample_host"]
    libraries = libraries.merge(
        stacked_samples[sel_col],
        left_on="archive_sample_accession",
        right_on="archive_accession",
    )
    select_libs = list(stacked_samples["archive_accession"])
    selected_libraries = libraries.query("archive_sample_accession in @select_libs")

    return selected_libraries


def get_filename(path_string: str, orientation: str) -> Tuple[str, str]:
    """
    Get Fastq Filename from download_links column

    Args:
        path_string(str): path to fastq files urls, comma separated
        orientation(str): [fwd | rev]
    Returns
        str: name of Fastq file
    """

    if ";" in path_string:
        fwd = Path(path_string.split(";")[0]).name
        rev = Path(path_string.split(";")[1]).name
    else:
        fwd = Path(path_string).name
        rev = "NA"
    if orientation == "fwd":
        return fwd
    elif orientation == "rev":
        return rev


def parse_to_mag(libraries):
    libraries["short_reads_1"] = libraries["download_links"].apply(
        get_filename, orientation="fwd"
    )
    libraries["short_reads_2"] = libraries["download_links"].apply(
        get_filename, orientation="rev"
    )
    libraries["short_reads_2"] = libraries["short_reads_2"].replace("NA", "")
    libraries["longs_reads"] = ""
    col2keep = [
        "archive_data_accession",
        "archive_sample_accession",
        "short_reads_1",
        "short_reads_2",
        "longs_reads",
    ]
    libraries = libraries[col2keep].rename(
        columns={
            "archive_data_accession": "sample",
            "archive_sample_accession": "group",
        }
    )
    return libraries


@st.cache_data
def prepare_eager_table(
    samples: pd.DataFrame,
    libraries: pd.DataFrame,
    table_name: str,
    supported_archives: Iterable[str],
) -> pd.DataFrame:
    """Prepare nf-core/eager tsv input table

    Args:
        sample (pd.dataFrame): selected samples table
        library (pd.dataFrame): library table
        table_name (str): Name of the table
        supported_archives (list): list of supported archives
    """

    libraries["Colour_Chemistry"] = libraries["instrument_model"].apply(
        get_colour_chemistry
    )

    libraries["UDG_Treatment"] = libraries.library_treatment.str.split(
        "-", expand=True
    )[0]

    libraries["R1"] = libraries["download_links"].apply(get_filename, orientation="fwd")

    libraries["R2"] = libraries["download_links"].apply(get_filename, orientation="rev")

    libraries["Lane"] = 0
    libraries["SeqType"] = where(libraries["library_layout"] == "SINGLE", "SE", "PE")
    libraries["BAM"] = "NA"
    if table_name == "ancientmetagenome-environmental":
        libraries["sample_host"] = "environmental"
    col2keep = [
        "sample_name",
        "archive_data_accession",
        "Lane",
        "Colour_Chemistry",
        "SeqType",
        "sample_host",
        "strand_type",
        "UDG_Treatment",
        "R1",
        "R2",
        "BAM",
    ]
    libraries = libraries[col2keep].rename(
        columns={
            "sample_name": "Sample_Name",
            "archive_data_accession": "Library_ID",
            "sample_host": "Organism",
            "strand_type": "Strandedness",
        }
    )

    return libraries


@st.cache_data
def prepare_mag_table(
    samples: pd.DataFrame,
    libraries: pd.DataFrame,
    table_name: str,
    supported_archives: Iterable[str],
) -> pd.DataFrame:
    """Prepare nf-core/mag tsv input table

    Args:
        sample (pd.dataFrame): selected samples table
        library (pd.dataFrame): library table
        table_name (str): Name of the table
        supported_archives (list): list of supported archives
    """

    # Create a DataFrame for "SINGLE" values
    single_libraries = libraries[libraries["library_layout"] == "SINGLE"]

    # Create a DataFrame for "PAIRED" values
    paired_libraries = libraries[libraries["library_layout"] == "PAIRED"]

    if not single_libraries.empty:
        single_libraries = parse_to_mag(single_libraries)
    if not paired_libraries.empty:
        paired_libraries = parse_to_mag(paired_libraries)

    return single_libraries, paired_libraries


@st.cache_data
def prepare_accession_table(
    samples: pd.DataFrame,
    libraries: pd.DataFrame,
    table_name: str,
    supported_archives: Iterable[str],
) -> pd.DataFrame:
    """Get accession lists for samples and libraries

    Args:
        samples (pd.dataFrame): selected samples table
        libraries (pd.dataFrame): library table
        table_name (str): Name of the table
        supported_archives (list): list of supported archives
    """

    # libraries = get_libraries(
    #     table_name=table_name,
    #     samples=samples,
    #     libraries=libraries,
    #     supported_archives=supported_archives,
    # )

    # Downloading with curl or aspera instead of fetchngs
    urls = set(libraries["download_links"])
    accessions = set(libraries["archive_data_accession"])
    links = set()
    for u in urls:
        for s in u.split(";"):
            links.add(s)
    dl_script_header = "#!/usr/bin/env bash\n"
    curl_script = (
        "\n".join([f"curl -L ftp://{l} -o {l.split('/')[-1]}" for l in links]) + "\n"
    )
    aspera_script = (
        "\n".join(
            [
                "ascp -QT -l 300m -P 33001 "
                "-i ${ASPERA_PATH}/etc/asperaweb_id_dsa.openssh "
                f"era-fasp@fasp.sra.ebi.ac.uk:{'/'.join(l.split('/')[1:])} ."
                for l in links
            ]
        )
        + "\n"
    )
    fasterq_dump_script = (
        "\n".join([f"fasterq-dump --split-files -p {a}" for a in accessions]) + "\n"
    )

    return {
        "df": libraries[["archive_data_accession", "download_sizes"]].drop_duplicates(),
        "curl_script": dl_script_header + curl_script,
        "aspera_script": dl_script_header + aspera_script,
        "fasterq_dump_script": dl_script_header + fasterq_dump_script,
    }


@st.cache_data
def prepare_taxprofiler_table(
    samples: pd.DataFrame,
    libraries: pd.DataFrame,
    table_name: str,
    supported_archives: Iterable[str],
) -> pd.DataFrame:
    """Prepare taxprofiler csv input table

    Args:
        sample (pd.dataFrame): selected samples table
        library (pd.dataFrame): library table
        table_name (str): Name of the table
        supported_archives (list): list of supported archives
    """

    libraries["fastq_1"] = libraries["download_links"].apply(
        get_filename, orientation="fwd"
    )

    libraries["fastq_2"] = libraries["download_links"].apply(
        get_filename, orientation="rev"
    )

    libraries["fastq_2"] = libraries["fastq_2"].replace("NA", "")

    libraries["fasta"] = ""

    libraries["instrument_model"] = where(
        libraries["instrument_model"]
        .str.lower()
        .str.contains("illumina|nextseq|hiseq|miseq"),
        "ILLUMINA",
        where(
            libraries["instrument_model"].str.lower().str.contains("torrent"),
            "ION_TORRENT",
            where(
                libraries["instrument_model"].str.lower().str.contains("helicos"),
                "HELICOS",
                where(
                    libraries["instrument_model"].str.lower().str.contains("bgiseq"),
                    "BGISEQ",
                    where(
                        libraries["instrument_model"].str.lower().str.contains("454"),
                        "LS454",
                        libraries["instrument_model"],
                    ),
                ),
            ),
        ),
    )

    col2keep = [
        "sample_name",
        "library_name",
        "instrument_model",
        "fastq_1",
        "fastq_2",
        "fasta",
    ]
    libraries = libraries[col2keep].rename(
        columns={
            "sample_name": "sample",
            "library_name": "run_accession",
            "instrument_model": "instrument_platform",
        }
    )

    return libraries


@st.cache_data
def prepare_aMeta_table(
    samples: pd.DataFrame,
    libraries: pd.DataFrame,
    table_name: str,
    supported_archives: Iterable[str],
) -> pd.DataFrame:
    """Prepare aMeta tsv input table

    Args:
        sample (pd.dataFrame): selected samples table
        library (pd.dataFrame): library table
        table_name (str): Name of the table
        supported_archives (list): list of supported archives
    """

    libraries["Colour_Chemistry"] = libraries["instrument_model"].apply(
        get_colour_chemistry
    )

    libraries["UDG_Treatment"] = libraries.library_treatment.str.split(
        "-", expand=True
    )[0]

    libraries["R1"] = libraries["download_links"].apply(get_filename, orientation="fwd")

    libraries["R2"] = libraries["download_links"].apply(get_filename, orientation="rev")

    libraries["Lane"] = 0
    libraries["SeqType"] = where(libraries["library_layout"] == "SINGLE", "SE", "PE")
    libraries["BAM"] = "NA"
    if table_name == "ancientmetagenome-environmental":
        libraries["sample_host"] = "environmental"
    col2keep = ["archive_data_accession", "R1"]
    libraries = libraries[col2keep].rename(
        columns={
            "archive_data_accession": "sample",
            "R1": "fastq",
        }
    )

    return libraries


@st.cache_data
def prepare_bibtex_file(samples: pd.DataFrame) -> str:
    dois = set()
    failed_dois = set()
    dois_set = set(list(samples["publication_doi"]))
    dois_set.add("10.1038/s41597-021-00816-y")
    for doi in dois_set:
        try:
            bibtex_str = doi2bib(doi)
            if len(bibtex_str) == 0:
                failed_dois.add(doi)
            else:
                dois.add(bibtex_str)
        except Exception as e:
            logger.info(e)
            pass
    # Print warning for DOIs that do not have an entry
    if len(failed_dois) > 0:
        st.warning(
            "Citation information could not be resolved for the "
            "following DOIs: " + ", ".join(failed_dois) + ". Please "
            "check how to cite these publications manually!"
        )
        logger.warning(
            "Citation information could not be resolved for the "
            "following DOIs: " + ", ".join(failed_dois) + ". Please "
            "check how to cite these publications manually!"
        )

    dois_string = "\n".join(list(dois))
    return dois_string


def is_merge_size_zero(
    samples: pd.DataFrame, library: pd.DataFrame, table_name: str
) -> bool:
    """
    Checks if intersection of samples and libraries table is not null

    Args:
        samples(pd.dataFrame): selected samples table
        libraries (pd.dataFrame): library table
        table_name (str): Name of the table
    """

    if samples.shape[0] == 0 or library.shape[0] == 0:
        return True
    stacked_samples = (
        samples["archive_accession"]
        .str.split(",", expand=True)
        .stack()
        .reset_index(level=0)
        .set_index("level_0")
        .rename(columns={0: "archive_accession"})
        .join(samples.drop("archive_accession", axis=1))
    )

    if table_name in [
        "ancientmetagenome-environmental",
    ]:
        sel_col = ["archive_accession"]
    else:
        sel_col = ["archive_accession", "sample_host"]
    library_selected = library.merge(
        stacked_samples[sel_col],
        left_on="archive_sample_accession",
        right_on="archive_accession",
    )

    if samples.shape[0] != 0 and library_selected.shape[0] == 0:
        return True
    return False



from pathlib import Path
import shutil
import os

from amdirt.convert import run_convert


def test_convert_only_sample_df(test_data_dir):
    assert (
        run_convert(
            samples=os.path.join(test_data_dir, "samples_test_convert.tsv"),
            libraries=None,
            table_name="ancientmetagenome-hostassociated",
            eager=True,
            fetchngs=True,
            ameta=True,
            output="test_files",
        )
        is None
    )
    shutil.rmtree("test_files")


def test_convert_only_libraries_df(test_data_dir):
    assert (
        run_convert(
            samples=os.path.join(test_data_dir, "samples_test_convert.tsv"),
            libraries=os.path.join(test_data_dir, "libraries_test_convert.tsv"),
            table_name="ancientmetagenome-hostassociated",
            eager=True,
            fetchngs=True,
            ameta=True,
            output="test_files",
        )
        is None
    )
    shutil.rmtree("test_files")



import pytest
import os
from amdirt.validate.domain import DatasetValidator, DFError
from jsonschema import validate
from jsonschema.exceptions import ValidationError as JSONValidationError


def test_DFError():
    err = DFError("test", "test", "test", 0, "test")
    assert err.to_dict() == {
        "Error": "test",
        "Source": "test",
        "Column": "test",
        "Row": "2",
        "Message": "test",
    }
    assert err.to_list() == ["test", "test", "test", "2", "test"]


@pytest.fixture
def schema():
    schema = {
        "type": "object",
        "properties": {
            "sample": {"type": "string"},
            "weight": {"type": "number"},
        },
    }
    return schema


@pytest.fixture
def valid_dataset():
    instance = {"sample": "bone", "weight": 10}
    return instance


@pytest.fixture
def invalid_dataset():
    instance = {"sample": 10, "weight": "bone"}
    return instance


def test_jsonschema_validator(valid_dataset, invalid_dataset, schema):

    assert validate(instance=valid_dataset, schema=schema) is None
    with pytest.raises(JSONValidationError):
        assert validate(instance=invalid_dataset, schema=schema)

def test_schema_parsing(test_data_dir):
    valid_dataset = os.path.join(test_data_dir, "valid.tsv")
    schema = os.path.join(test_data_dir, "schema.json")
    invalid_schema_too_many_types = os.path.join(test_data_dir, "invalid_schema_too_many_types.json")
    invalid_schema_parsing = os.path.join(test_data_dir, "invalid_schema_parsing.json")

    assert DatasetValidator(schema=schema, dataset=valid_dataset)

    assert DatasetValidator(schema=invalid_schema_too_many_types, dataset=valid_dataset).parsing_ok is False

    assert DatasetValidator(schema=invalid_schema_parsing, dataset=valid_dataset).parsing_ok is False


def test_dataset_parsing(test_data_dir):
    valid_dataset = os.path.join(test_data_dir, "valid.tsv")
    invalid_dataset_too_many_columns = os.path.join(
        test_data_dir, "invalid_too_many_columns.tsv"
    )
    invalid_dataset_invalid_dtype_in_col = os.path.join(
        test_data_dir, "invalid_dtype_in_col.tsv"
    )
    schema = os.path.join(test_data_dir, "schema.json")

    assert DatasetValidator(schema=schema, dataset=valid_dataset)

    assert DatasetValidator(schema, invalid_dataset_too_many_columns).parsing_ok is False

    assert DatasetValidator(schema, invalid_dataset_invalid_dtype_in_col).parsing_ok is False


def test_column_name(test_data_dir):
    valid_dataset = os.path.join(test_data_dir, "valid.tsv")
    invalid_column_name = os.path.join(test_data_dir, "invalid_column_name.tsv")
    schema = os.path.join(test_data_dir, "schema.json")

    valid = DatasetValidator(schema=schema, dataset=valid_dataset)
    invalid = DatasetValidator(schema=schema, dataset=invalid_column_name)

    assert valid.check_columns() is True

    assert invalid.check_columns() is False


def test_duplicate_row(test_data_dir):
    valid_dataset = os.path.join(test_data_dir, "valid.tsv")
    invalid_dataset = os.path.join(test_data_dir, "invalid_duplicate_row.tsv")
    schema = os.path.join(test_data_dir, "schema.json")

    validator_valid = DatasetValidator(schema=schema, dataset=valid_dataset)
    validator_invalid = DatasetValidator(schema, invalid_dataset)

    assert validator_valid.check_duplicate_rows() is True
    assert validator_invalid.check_duplicate_rows() is False


def test_schema_validation(test_data_dir):
    valid_dataset = os.path.join(test_data_dir, "valid.tsv")
    invalid_dataset = os.path.join(test_data_dir, "invalid_schema.tsv")
    schema = os.path.join(test_data_dir, "schema.json")

    validator_valid = DatasetValidator(schema=schema, dataset=valid_dataset)

    assert validator_valid.validate_schema() is True

    assert validator_valid.check_columns() is True

    validator_invalid = DatasetValidator(schema, invalid_dataset)

    assert validator_invalid.validate_schema() is False

    with pytest.raises(SystemExit):
        validator_invalid.to_rich()

    with pytest.raises(SystemExit):
        validator_invalid.to_markdown()



from setuptools import setup, find_packages
import codecs
import os.path


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


setup(
    name="amdirt",
    version=get_version("amdirt/__init__.py"),
    description="amdirt: AncientMetagenomeDir Toolkit",
    long_description=open("README.md").read(),
    url="https://github.com/SPAAM-community/amdirt",
    long_description_content_type="text/markdown",
    license="GNU-GPLv3",
    python_requires=">=3.9",
    install_requires=[
        "click",
        "pandas",
        "tabulate",
        "jsonschema",
        "rich",
        "streamlit<=1.35.0",
        "streamlit-aggrid",
        "numpy",
        "requests",
        "colorlog"
    ],
    packages=find_packages(exclude="test"),
    entry_points={"console_scripts": ["amdirt = amdirt.cli:cli"]},
    package_dir={"amdirt": "amdirt"},
    package_data={"amdirt.assets": ["tables.json"]},
)



from amdirt.validate.application import AMDirValidator
import warnings

def run_validation(
    dataset,
    schema,
    schema_check,
    line_dup,
    columns,
    doi,
    multi_values,
    online_archive,
    remote,
    markdown,
    verbose,
):
    if not verbose:
        warnings.filterwarnings("ignore")
    v = AMDirValidator(schema, dataset)
    if schema_check and v.parsing_ok:
        v.validate_schema()
    if line_dup and v.parsing_ok:
        v.check_duplicate_rows()
    if columns and v.parsing_ok:
        v.check_columns()
    if doi and v.parsing_ok:
        v.check_duplicate_dois()
    if multi_values and v.parsing_ok: 
        v.check_multi_values(column_names=multi_values)
    if online_archive and v.parsing_ok:
        v.check_sample_accession(remote=remote)
    if markdown:
        v.to_markdown()
    else:
        v.to_rich()



from sys import path
import os
from amdirt.core import (
    prepare_accession_table,
    prepare_bibtex_file,
    prepare_eager_table,
    prepare_mag_table,
    prepare_aMeta_table,
    is_merge_size_zero,
    prepare_taxprofiler_table,
    get_libraries,
    get_remote_resources,
    get_json_path,
)
from amdirt.validate import AMDirValidator
from amdirt.validate.exceptions import DatasetValidationError
from json import load
from amdirt.core import logger
import pandas as pd
import warnings


def run_convert(
    samples,
    libraries,
    table_name,
    tables=None,
    output=".",
    bibliography=False,
    librarymetadata=False,
    curl=False,
    aspera=False,
    eager=False,
    fetchngs=False,
    sratoolkit=False,
    ameta=False,
    taxprofiler=False,
    mag=False,
    verbose=False,
):
    """Run the amdirt conversion application to input samplesheet tables for different pipelines

    Args:
        samples (str): Path to AncientMetagenomeDir filtered samples tsv file
        libraries(str): Optional path to AncientMetagenomeDir pre-filtered libraries tsv file
        table_name (str): Name of the table of the table to convert
        tables (str): Path to JSON file listing tables
        output (str): Path to output table. Defaults to "."
    """
    os.makedirs(output, exist_ok=True)

    if not verbose:
        warnings.filterwarnings("ignore")
    supported_archives = ["ENA", "SRA"]

    # Validate input table
    if tables is None:
        remote_resources = get_remote_resources()
    else:
        with open(tables, "r") as f:
            remote_resources = load(f)

    if table_name not in remote_resources["samples"]:
        raise ValueError(f"{table_name} not found in AncientMetagenomeDir file")
    if not verbose:
        warnings.filterwarnings("ignore")

    schema = remote_resources[f"samples_schema"][table_name]
    dataset_valid = list()
    v = AMDirValidator(schema, samples)
    dataset_valid.append(v.parsing_ok)
    if v.parsing_ok:
        dataset_valid.append(v.validate_schema())
        dataset_valid.append(v.check_duplicate_rows())
        dataset_valid.append(v.check_columns())

    dataset_valid = all(dataset_valid)
    if dataset_valid is False:
        v.to_rich()
        raise DatasetValidationError("Input sample dataset is not valid")
    else:
        logger.info("Input sample dataset is valid")
        samples = pd.read_csv(samples, sep="\t")
        remote_libraries = pd.read_csv(
            remote_resources["libraries"][table_name], sep="\t"
        )

    if not libraries:
        selected_libraries = get_libraries(
            samples=samples,
            libraries=remote_libraries,
            table_name=table_name,
            supported_archives=supported_archives,
        )
    else:
        schema = remote_resources[f"libraries_schema"][table_name]
        dataset_valid = list()
        v = AMDirValidator(schema, libraries)
        dataset_valid.append(v.parsing_ok)
        if v.parsing_ok:
            dataset_valid.append(v.validate_schema())
            dataset_valid.append(v.check_duplicate_rows())
            dataset_valid.append(v.check_columns())

        dataset_valid = all(dataset_valid)
        if dataset_valid is False:
            v.to_rich()
            raise DatasetValidationError("Input libraries dataset is not valid")
        else:
            logger.info("Input libraries dataset is valid")
            libraries = pd.read_csv(libraries, sep="\t")
            selected_libraries = get_libraries(
                samples=samples,
                libraries=libraries,
                table_name=table_name,
                supported_archives=supported_archives,
            )

    accession_table = prepare_accession_table(
        samples=samples,
        libraries=selected_libraries,
        table_name=table_name,
        supported_archives=supported_archives,
    )

    logger.warning(
        "We provide no warranty to the accuracy of the generated input sheets."
    )

    if bibliography == True:
        bibfile = f"{output}/AncientMetagenomeDir_bibliography.bib"
        logger.info(f"Writing Bibtex citation file to {bibfile}")
        with open(bibfile, "w") as fw:
            fw.write(prepare_bibtex_file(samples))

    if table_name in ["ancientmetagenome-environmental"]:
        col_drop = ["archive_accession"]
    else:
        col_drop = ["archive_accession", "sample_host"]

    if librarymetadata == True:
        tbl_file = f"{output}/AncientMetagenomeDir_filtered_libraries.tsv"
        logger.info(f"Writing filtered libraries table to {tbl_file}")
        librarymetadata = selected_libraries.drop(col_drop, axis=1)
        librarymetadata.to_csv(
            tbl_file,
            sep="\t",
            index=False,
        )

    if curl == True:
        dl_file = f"{output}/AncientMetagenomeDir_curl_download_script.sh"
        logger.info(f"Writing curl download script to {dl_file}")
        with open(dl_file, "w") as fw:
            fw.write(accession_table["curl_script"])

    if aspera == True:
        dl_file = f"{output}/AncientMetagenomeDir_aspera_download_script.sh"
        logger.info(f"Writing Aspera download script to {dl_file}")
        logger.warning(
            "You will need to set the ${ASPERA_PATH} environment variable. See https://amdirt.readthedocs.io for more information."
        )
        with open(dl_file, "w") as fw:
            fw.write(accession_table["aspera_script"])

    if fetchngs == True:
        dl_file = f"{output}/AncientMetagenomeDir_nf_core_fetchngs_download_script.sh"
        logger.info(f"Writing nf-core/fetchngs table to {dl_file}")
        accession_table["df"]["archive_data_accession"].to_csv(
            dl_file,
            sep="\t",
            header=False,
            index=False,
        )
    if sratoolkit == True:
        dl_file = f"{output}/AncientMetagenomeDir_sratoolkit_download_script.sh"
        logger.info(f"Writing sratoolkit/fasterq-dump download script to {dl_file}")
        with open(dl_file, "w") as fw:
            fw.write(accession_table["fasterq_dump_script"])

    if eager == True:
        tbl_file = f"{output}/AncientMetagenomeDir_nf_core_eager_input_table.tsv"
        logger.info(f"Writing nf-core/eager table to {tbl_file}")
        eager_table = prepare_eager_table(
            samples=samples,
            libraries=selected_libraries,
            table_name=table_name,
            supported_archives=supported_archives,
        )
        eager_table.to_csv(
            tbl_file,
            sep="\t",
            index=False,
        )

    if taxprofiler == True:
        tbl_file = f"{output}/AncientMetagenomeDir_nf_core_taxprofiler_input_table.csv"
        logger.info(f"Writing nf-core/taxprofiler table to {tbl_file}")
        taxprofiler_table = prepare_taxprofiler_table(
            samples=samples,
            libraries=selected_libraries,
            table_name=table_name,
            supported_archives=supported_archives,
        )
        taxprofiler_table.to_csv(
            tbl_file,
            header=False,
            index=False,
        )

    if ameta == True:
        tbl_file = f"{output}/AncientMetagenomeDir_aMeta_input_table.tsv"
        logger.info(f"Writing aMeta table to {tbl_file}")
        logger.warning(
            "aMeta does not support pairs. You must manually merge pair-end data before using samplesheet."
        )
        aMeta_table = prepare_aMeta_table(
            samples=samples,
            libraries=selected_libraries,
            table_name=table_name,
            supported_archives=supported_archives,
        )

        aMeta_table.to_csv(
            tbl_file,
            sep="\t",
            index=False,
        )

    if mag == True:
        logger.info("Preparing nf-core/mag table(s)")
        mag_table_single, mag_table_paired = prepare_mag_table(
            samples=samples,
            libraries=selected_libraries,
            table_name=table_name,
            supported_archives=supported_archives,
        )
        if not mag_table_single.empty:
            mag_tbl_single_file = (
                f"{output}/AncientMetagenomeDir_nf_core_mag_input_single_table.csv"
            )
            logger.info(
                f"Writing nf-core/mag single-end table to {mag_tbl_single_file}"
            )
            mag_table_single.to_csv(
                mag_tbl_single_file,
                index=False,
            )
        if not mag_table_paired.empty:
            mag_tbl_paired_file = (
                f"{output}/AncientMetagenomeDir_nf_core_mag_input_paired_table.csv"
            )
            logger.info(
                f"Writing nf-core/mag paired-end table to {mag_tbl_paired_file}"
            )
            mag_table_paired.to_csv(
                mag_tbl_paired_file,
                index=False,
            )
